/*
 * Paged Translation Table
 * 
 * Aug 93    Dereference functions have been restructured.
 * Jul 93    Portable Implementation and Integration - Mustafa
 * May 93    Bug Fixes - Mustafa
 * May 93    First implementation for i860 - Raja
 */

#define TT_ERROR   -1

#include <stdio.h>
#include <assert.h>
#include "parti.h"




/*
 * tableGetReplication
 *
 * Returns the replication factor of the translation table
 *
 */
float tableGetReplication(table)
     TTABLE *table;
{
  if(table->type == DISTRIBUTED)
    return(table->tabptr->repFactor);
  else
    return((float)TT_ERROR); 
}
     





/*
 * tableGetPageSize
 *
 * Returns the number of elements stored in one block.
 *
 */
int tableGetPageSize(table)
     TTABLE *table;
{
  if(table->type == DISTRIBUTED)
    return(table->tabptr->pageSize);
  else
    return(TT_ERROR);
}
     
static int
intcompare(x, y)
int *x, *y;
{
 return (*x-*y);
}




/*
 * tableGetIndices
 *
 * Returns a pointer to the global indices assigned to the calling processor 
 *
 * Added support  for regular translation table - shamik & pravi
 */
intPtr tableGetIndices(table)
     TTABLE *table;
{
  DSTABLEPTR dtable;
  REGTABLEPTR rtable; 
  int extra, start;
  intPtr indx, myIndices;
  SCHED *sr, *schedule_proc();
  int  nIndex, i, j, cnt, newNdata;

  if (table->type == DISTRIBUTED) 
    {
      dtable = table->tabptr;
      
      nIndex = dtable->pageSize*dtable->numPages;
      indx = (intPtr) malloc(sizeof(int)*nIndex);

      if (tableGetReplication(table) >= 1.0) {
        for (cnt = j=0; j<dtable->pageSize; j++)
            if (dtable->Lproc[j] == PARTI_mynode)
              indx[cnt++] = j;
        
        return (indx);
      }
      for (cnt=i=0; i<dtable->numPages; i++)
        for (j=0; j<dtable->pageSize; j++)
	  indx[cnt++] = PARTI_mynode*dtable->numPages + i*dtable->numPages+j;
      
      sr = schedule_proc(dtable->Lproc, nIndex, &newNdata, 1);
      myIndices = (intPtr) malloc(sizeof(int)*newNdata);
      PARTI_scatter_append(sr, indx, myIndices, sizeof(int));
      qsort (myIndices, newNdata, sizeof(int), intcompare);
      free(indx);
      return(myIndices);
    }
  else if ( table->type ==  REGULAR ) 
    {
      rtable     = table->regptr ;
      nIndex     = rtable->block_size;
      myIndices  = (intPtr) malloc(sizeof(int)*nIndex);
      if ( rtable->dist_type == BLOCKED ) 
	{
	  extra = ( rtable->data_size % PARTI_numnodes );
	  if  (  PARTI_mynode < extra ) 
	    {
	      start = PARTI_mynode * nIndex ;
	    }
	  else 
	    {
	      start = (extra *(nIndex+1)) + (PARTI_mynode-extra)*nIndex;
	    }

	  for (j=0; j<rtable->block_size; j++) 
	    {
	      myIndices[j] = start + j;
	    }

	  return(myIndices);
	}
      else if ( rtable->dist_type == CYCLIC ) 
	{
	  fprintf(stderr,"Error: tableGetIndices not supported for cyclic");
	} 
    }
}






/*
 * tableGetIndexCount
 *
 * Returns the number of table entries assigned to calling processor
 */
int tableGetIndexCount(table)
     TTABLE *table;
{
  /* consistency checks */
  if( table->type == DISTRIBUTED )
    return(table->tabptr->nMyIndices);
  else if (table->type == REGULAR)
    return(table->regptr->block_size);
  else
    fprintf(stderr, "Invalid Translation table type"), exit(1);
}



/*
 * build_reg_translation_table
 *
 * Translation table building function for the regular distributions.
 * No explicit data is stored. Only the (regular) information associated
 * with the dereference formulae is stored.
 */
TTABLE  *build_reg_translation_table (  dist_type, data_size )
     int  dist_type, data_size; 
{
  TTABLE    *trans_table;
  REG_TABLE *table;
  int  block_size , extra ;
  
  trans_table = (TTABLE *) malloc(sizeof(TTABLE)); 
  trans_table->type   = REGULAR ;
  trans_table->regptr = table =(REG_TABLE *)malloc(sizeof(REG_TABLE));
  trans_table->tabptr = NULL;
  table->dist_type    = dist_type ;
  table->data_size    = data_size;

  block_size = data_size / PARTI_numnodes ;
  extra      = data_size % PARTI_numnodes;
  if ( extra != 0 && ( extra > PARTI_mynode ))  
     block_size++ ;

  table->block_size   = block_size; 
  return ( trans_table) ;
}




/*
 * mkDistTable
 *
 * Builds a Distributed translation table.
 * This is a wrapper for BuildDistTable.
 */
TTABLE  *mkDistTable(index_array, ndata, replication, pageSize)
     intPtr index_array;
     float replication;
     int   ndata, pageSize;
{
  TTABLE  *trans_table;
  DSTABLE *table, *BuildReplicated();;

  trans_table         = (TTABLE *) malloc(sizeof(TTABLE));
  trans_table->type   = DISTRIBUTED ;  
  trans_table->regptr = NULL ; 
  if (replication >= 1.0)
    trans_table->tabptr = BuildReplicated(index_array, ndata);
  else
    trans_table->tabptr = BuildDistTable(index_array, ndata, replication, pageSize);
  
  return(trans_table);
}




void    free_table(trans_table)
     TTABLE *trans_table;
{
  DSTABLE *table;

  if (trans_table->type == REGULAR ) 
    {
      free( trans_table->regptr );
      free( trans_table);
    }
  else  
    {
      table = trans_table->tabptr ;
      DestroyDistTable(table);
      free( trans_table );
    }
}





/*
 * initDistTable
 *
 * initializes the translation table in the irregular mode with the
 * values provided in the plist. The translation table will contain
 * only one page in each processor and will be completely distributed.
 */
DSTABLE *InitDistTable(plist, ndata)
     int *plist;   /* the data array used to fill in the TT entries */ 
     int ndata;    /* size of plist */
{
  DSTABLE *Table;
  int Nprocs, i, *proc, totalPages;

  Nprocs = PARTI_numnodes;
  Table = (DSTABLE *) malloc(sizeof(DSTABLE));
  Table->numPages = 1;
  totalPages = Table->numPages * Nprocs;
  Table->pageTable = (PAGEPTR *)malloc(sizeof(PAGEPTR)*totalPages);
  for(i=0;i<(Table->numPages*Nprocs);i++)
    Table->pageTable[i] = NULL;
  Table->pageSize = ndata;
  
  Table->pageTable[PARTI_mynode] = (PAGEPTR) malloc(sizeof(PAGE));
  proc = (int *) malloc(sizeof(int)*ndata);
  for(i=0; i<ndata; i++)
    proc[i] = plist[i];

  Table->Lproc = NULL;
  Table->Loffset = NULL;
  Table->pageTable[PARTI_mynode]->Offset = NULL;
  Table->pageTable[PARTI_mynode]->Proc = proc;

  return(Table);
}





DSTABLE *BuildReplicated(IndexArray, nindex)
     int *IndexArray, nindex;
{
  DSTABLE *table;
  intPtr wrk, MyProc, MyOffset;
  int MaxIndex,i,nd;

  /* find the largest index to be addressed */
  MaxIndex = 0;
  for(i=0;i<nindex;i++)
    {
      MaxIndex = (IndexArray[i]>MaxIndex) ? IndexArray[i] : MaxIndex;
    }  
  PARTI_gihigh(&MaxIndex, 1, &i);
  nd = nindex;
  PARTI_gisum(&nd, 1, &i);
  MaxIndex = (MaxIndex >= nd) ? MaxIndex : nd;

  table = (DSTABLE *)malloc(sizeof(DSTABLE));
  table->repFactor = 1.0;
  table->pageSize  = MaxIndex;

  MyOffset = (int *) malloc(sizeof(int)*MaxIndex);
  MyProc   = (int *) malloc(sizeof(int)*MaxIndex);
  wrk      = (int *) malloc(sizeof(int)*MaxIndex);

  for (i=0; i<MaxIndex; i++)
    {
      MyProc[i] = MyOffset[i] = 0;
    }

  for (i=0; i<nindex; i++)
    {
      MyProc[IndexArray[i]] = PARTI_mynode;
      MyOffset[IndexArray[i]] = i;
    }

  PARTI_gisum(MyProc, MaxIndex, wrk);
  PARTI_gisum(MyOffset, MaxIndex, wrk);

  table->Loffset = MyOffset;
  table->Lproc   = MyProc;

  free(wrk);

  return(table);
}





/*
 * BuildDistTable
 *
 * Builds distributed translation tables. Each index is stored
 * along with the processor and the local offset.
 * nb: Slightly modified version of Raja's initial implementation. 
 */
DSTABLE  *BuildDistTable(IndexArray,Ndata,Memp,blkSize)
     int Ndata,*IndexArray, blkSize;
     float Memp;
{
  int MaxIndex,TableSize,i,*Offset,*Local,*Proc;
  SCHED *sr;
  DSTABLE *Table;
  PAGEPTR LocalChunk;
  int pgsize, totalpgs, localpgs;
  int Nprocs,Myproc,replpgs;
  int *MyOffset, *MyProc;

  int j,Procnum,Extra,NumProcs;
  int loffset,count,LastBlock;
  int poffset,counter,Length;
  int full,fullcounter,*Local1,*Proc1;
  int wrk, nd;

  /* correctness checks */
  assert( Memp >= 0.0 && Memp <= 1.0);
  assert( blkSize != 0);

  /* get the environment */
  Nprocs = PARTI_numnodes;
  Myproc = PARTI_mynode;

  /* find the largest index to be addressed */
  MaxIndex = 0;
  for(i=0;i<Ndata;i++)
    {
      MaxIndex = (IndexArray[i]>MaxIndex) ? IndexArray[i] : MaxIndex;
    }  
  PARTI_gihigh(&MaxIndex, 1, &i);
  nd = Ndata;
  PARTI_gisum(&nd, 1, &i);
  MaxIndex = (MaxIndex >= nd) ? MaxIndex : nd;

  /* Find # of pages, pagesize, etc */
  if (blkSize>=0)
    {
      pgsize   = blkSize;
      totalpgs = (MaxIndex%pgsize) ? MaxIndex/pgsize+1 : MaxIndex/pgsize;
    }
  else
    {
      pgsize   = MaxIndex / Nprocs + 1;
      totalpgs = Nprocs;
    }
  if (totalpgs%Nprocs)
    localpgs = totalpgs / Nprocs + 1;
  else
    localpgs  = totalpgs / Nprocs;
  totalpgs = localpgs * PARTI_numnodes;
  replpgs = (int)(Memp * (totalpgs-localpgs));

  /* Create the table structure */
  Table = (DSTABLE *)malloc(sizeof(DSTABLE));
  Table->pageTable = (PAGEPTR*)malloc(sizeof(PAGEPTR)*(localpgs*Nprocs));
  for(i=0;i<totalpgs;i++)
    Table->pageTable[i] = NULL;
  Table->pageSize = pgsize;
  Table->Replication = replpgs;
  Table->numPages = localpgs;
  Table->nMyIndices = Ndata;
  Table->repFactor = Memp;

  MyOffset = (int *) malloc(sizeof(int)*(pgsize*localpgs));
  MyProc   = (int *) malloc(sizeof(int)*(pgsize*localpgs));
  
  /* Temporary arrays for doing the data movement */
  Offset = (int *) malloc(sizeof(int)*Ndata);
  Local = (int *) malloc(sizeof(int)*Ndata);
  Proc = (int *) malloc(sizeof(int)*Ndata);
  
  /* Find out in which chunks this will belong */
  for(i=0;i<Ndata;i++){
    Offset[i] = i;
    Proc[i] = IndexArray[i] / (pgsize*localpgs);
    Local[i] = IndexArray[i] % (pgsize*localpgs);
  }              

  /* form schedule to send local stuff to right chunk */
  sr = schedule(Local,Proc,Ndata,1);
  iscatter(Offset,MyOffset,sr);
  
  for(i=0;i<Ndata;i++){
    Offset[i] = Myproc;
  }  
  iscatter(Offset,MyProc,sr);
  free_sched(sr);
  
  Table->Loffset = MyOffset;
  Table->Lproc   = MyProc;
  
  loffset = Myproc*localpgs;
  count = 0;
  for(i=0;i<localpgs;i++)
    {
      LocalChunk = (PAGEPTR)malloc(sizeof(PAGE));
      LocalChunk->Offset = &MyOffset[count];
      LocalChunk->Proc   = &MyProc[count];
      count += pgsize;
      Table->pageTable[loffset+i] = LocalChunk;
    }
  free(Offset); free(Local); free(Proc);

  /* move in as many pages as possible */
  Proc = (intPtr) malloc(sizeof(int)*Table->Replication);
  localpgs = Table->numPages * Myproc + Table->numPages;
  for (count=0,i=1; i<=Table->Replication; i++,count++)
    Proc[count] = (localpgs + i) % totalpgs;
  getPage(Table, Proc, Table->Replication);
  Table->Replication = 0;
  free(Proc);

  return(Table);
}



/*
 * remap_table
 *  
 * Updates the entries in the translation table to reflect the
 * recent data remapping operation. 
 *
 */
int remap_table(oldtable, newIndex, nindex)
     TTABLEPTR oldtable;
     intPtr newIndex;
     int    nindex;
{
  DSTABLE *Table,*oldt;
  PAGEPTR LocalChunk;
  intPtr  MyProc, MyOffset, invalid, isum, wrk, proc, iptr, newValues;
  int     lclsize, myBase, limit, count, totpgs, psize, j;
  int     pageNo, pageOffs;
  int     ptecnt, i, to, from;

  if (oldtable->type != DISTRIBUTED)
    return;

  oldt = oldtable->tabptr;
  Table = oldt;
  if (Table->repFactor >= 1.0)
    {
      for (i=0; i<Table->pageSize; i++)
	Table->Lproc[i] = Table->Loffset[i] = 0;
      for (i=0; i<nindex; i++)
	{
	  Table->Lproc[newIndex[i]] = PARTI_mynode;
	  Table->Loffset[newIndex[i]] = i;
	}
      
      wrk = (int *)malloc(sizeof(int)*Table->pageSize);
      PARTI_gisum(Table->Lproc, Table->pageSize, wrk);
      PARTI_gisum(Table->Loffset, Table->pageSize, wrk);
      
      free(wrk);
    }
  else
    {
      Table = oldt;
      totpgs = PARTI_numnodes * oldt->numPages;
      psize = sizeof(int)*oldt->pageSize;
      
      /* update the changed pages now */
      newValues = (intPtr) malloc(totpgs*psize);
      for (j=0; j<PARTI_numnodes;j++)
	{
	  to = (PARTI_mynode+j) % PARTI_numnodes;
	  if (j)
	    {
	      PARTI_csend(TABLE_REMAP, newIndex, sizeof(int)*nindex, to, 0);
	      PARTI_crecv(TABLE_REMAP, newValues, totpgs*psize,PARTI_ANY_NODE,0 );
	      limit = PARTI_infocount()/sizeof(int);
	      from = PARTI_infonode();
	      iptr = newValues;
	    }
	  else
	    {
	      iptr = newIndex;
	      limit = nindex;
	      from = PARTI_mynode;
	    }
	  
	  for (i=0; i<limit; i++)
	    {
	      pageNo = iptr[i] / Table->pageSize;
	      if (Table->pageTable[pageNo] != NULL)
		{
		  pageOffs = iptr[i] % Table->pageSize;
		  Table->pageTable[pageNo]->Offset[pageOffs] = i;
		  Table->pageTable[pageNo]->Proc[pageOffs] = from;
		}
	    }
	}
      
      free(newValues);
    }
}







/*
 * DestroyDistTable
 *
 * Reclaims the memory space for a translation table.
 */
void DestroyDistTable( tableptr )
     DSTABLE *tableptr;
{
  PAGEPTR  chnkptr;
  int      nslots, mystart, myend;
  int i;
  
  if (tableptr == NULL)
    {
      printf("\nTrying to destroy nonexistent translation table...");
      return;
    }
  
  nslots  = tableptr->numPages * PARTI_numnodes;
  mystart = tableptr->numPages * PARTI_mynode;
  myend   = tableptr->numPages + mystart;

  for (i=0; i<nslots; i++)
    {
      chnkptr = tableptr->pageTable[i];
      if (chnkptr != NULL)
	{
	  if (i<mystart || i >= myend)
	    {
	      if (chnkptr->Proc==NULL  ||  chnkptr->Offset==NULL)
		printf("\n NULL proc or Offset ptr in page table");

	      free(chnkptr->Proc);
	      free(chnkptr->Offset);
	    }
	  free(chnkptr);
	}
    }

  if (tableptr->Loffset != NULL)
    free(tableptr->Loffset);
  if (tableptr->Lproc != NULL)
    free(tableptr->Lproc);

  free(tableptr);
}





/*
 * freePage
 *
 * Destroy the memory allocated for a dynamic (mobile) page.
 *
 */
PAGEPTR freePage( pagePtr )
     PAGEPTR pagePtr;
{
  free(pagePtr->Offset);
  free(pagePtr->Proc);
  free(pagePtr);
  
  return(NULL);
}     







